/* client2.c */

#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>
#include <stdio.h>

typedef enum { FALSE, TRUE } bool;

typedef struct
{
  char          name[128+1];
  SQLSMALLINT   nameLength;
  SQLSMALLINT   dataType;
  SQLUINTEGER   fieldLength;
  SQLSMALLINT   scale;
  SQLSMALLINT   nullable;
  SQLINTEGER    displaySize;
  int           headerLength;
  SQLINTEGER    resultLength;
  char        * value;
} resultField;

static void printResultSet( SQLHSTMT stmt );

static bool SQL_OK( SQLRETURN result )
{
  if( result == SQL_SUCCESS || result == SQL_SUCCESS_WITH_INFO )
    return( TRUE );
  else
    return( FALSE );
}

static bool printErrors( SQLHENV  envHandle,
                         SQLHDBC  conHandle,
                         SQLHSTMT stmtHandle )
{
  SQLRETURN   result;
  SQLCHAR     sqlState[6];
  SQLINTEGER  nativeError;
  SQLSMALLINT requiredLength;
  SQLCHAR     messageText[SQL_MAX_MESSAGE_LENGTH+1];

  do
  {
    result = SQLError( envHandle,
                       conHandle,
                       stmtHandle,
                       sqlState,
                       &nativeError,
                       messageText,
                       sizeof( messageText ),
                       &requiredLength );

    if( SQL_OK( result ))
      {
        printf( "SQLState     = %s\n", sqlState );
        printf( "Native error = %d\n", nativeError );
        printf( "Message text = %s\n", messageText );
      }
  } while( SQL_OK( result ));
}

static void executeStmt( SQLHDBC con, char * stmtText )
{
  SQLHSTMT  stmt;

  SQLAllocHandle( SQL_HANDLE_STMT, con, &stmt );

  if( SQL_OK( SQLExecDirect( stmt, stmtText, SQL_NTS )))
    printResultSet( stmt );
  else
    printErrors( SQL_NULL_HENV, SQL_NULL_HDBC, stmt );
}

static void printResultSet( SQLHSTMT stmt )
{
  SQLSMALLINT   i;
  SQLSMALLINT   columnCount;
  resultField * fields;

  //  First, examine the metadata for the 
  //  result set so that we know how much
  //  room we need for each column.

  SQLNumResultCols( stmt, &columnCount );

  fields = (resultField *)calloc( columnCount+1, 
				  sizeof( resultField ));

  for( i = 1; i <= columnCount; i++ )
  {
    SQLDescribeCol( stmt,
		    i,
		    fields[i].name,
		    sizeof( fields[i].name ),
		    &fields[i].nameLength,
		    &fields[i].dataType,
		    &fields[i].fieldLength,
		    &fields[i].scale,
		    &fields[i].nullable );

    SQLColAttribute( stmt, 
		     i,
		     SQL_DESC_DISPLAY_SIZE,
		     NULL,
		     0,
		     NULL,
		     &fields[i].displaySize );


    fields[i].value = (char *)malloc( fields[i].displaySize + 1 );

    if( fields[i].nameLength > fields[i].displaySize )
      fields[i].headerLength = fields[i].nameLength;
    else
      fields[i].headerLength = fields[i].displaySize;
  }

  //  Now print out the column headers

  for( i = 1; i <= columnCount; i++ )
  {
    printf( "%-*s ", fields[i].headerLength, fields[i].name );
  }
  printf( "\n" );

  //  Now fetch and display the results...

  while( SQL_OK( SQLFetch( stmt )))
  {
    for( i = 1; i <= columnCount; i++)
    {
      SQLRETURN result;

      result = SQLGetData( stmt,
			   i,
			   SQL_C_CHAR,
			   fields[i].value,
			   fields[i].displaySize,
			   &fields[i].resultLength );

      if( fields[i].resultLength == SQL_NULL_DATA )
	printf( "%-*s ", fields[i].headerLength, "" );
      else
	printf( "%-*s ", fields[i].headerLength, fields[i].value );
    }
    printf( "\n" );
  }
  
  for( i = 1; i <= columnCount; i++ )
    free( fields[i].value );

  free( fields );

}

int main( int argc, char * argv[] )
{
  SQLRETURN   res;
  SQLHENV     env;
  SQLHDBC     con;
  SQLCHAR     fullConnectStr[SQL_MAX_OPTION_STRING_LENGTH];
  SQLSMALLINT requiredLength;

  res = SQLAllocHandle( SQL_HANDLE_ENV, SQL_NULL_HANDLE, &env );

  if( SQL_OK( res ))
  {
    res = SQLSetEnvAttr( env, 
                         SQL_ATTR_ODBC_VERSION, 
                         (SQLPOINTER)SQL_OV_ODBC2, 
                         0 );  
    if( !SQL_OK( res ))
    {
      printErrors( env, SQL_NULL_HDBC, SQL_NULL_HSTMT );
      exit( -1 );
    }

    res = SQLAllocHandle( SQL_HANDLE_DBC, env, &con );
    if( !SQL_OK( res ))
    {
      printErrors( env, SQL_NULL_HDBC, SQL_NULL_HSTMT );
      exit( -2 );
    }

    res = SQLDriverConnect( con, 
                            (SQLHWND)NULL,
                            argv[1], SQL_NTS,
                            fullConnectStr, 
                            sizeof( fullConnectStr ),
                            &requiredLength,
                            SQL_DRIVER_NOPROMPT );

                              
    if( !SQL_OK( res ))
    {
      printErrors( SQL_NULL_HENV, con, SQL_NULL_HSTMT );
      exit( -3 );
    }

    printf( "connection ok\n" );

    executeStmt( con, argv[2] );

    res = SQLDisconnect( con );
    if( !SQL_OK( res ))
    {
      printErrors( SQL_NULL_HENV, con, SQL_NULL_HSTMT );
      exit( -4 );
    }

    res = SQLFreeHandle( SQL_HANDLE_DBC, con );
    if( !SQL_OK( res ))
    {
      printErrors( SQL_NULL_HENV, con, SQL_NULL_HSTMT );
      exit( -5 );
    }

    res = SQLFreeHandle( SQL_HANDLE_ENV, env );
    if( !SQL_OK( res ))
    {
      printErrors( env, SQL_NULL_HDBC, SQL_NULL_HSTMT );
      exit( -6 );
    }
  }

  exit( 0 );

}
