/* client4.c */

#include <stdio.h>
#include <string.h>
#include <libpq-fe.h>
#include <libpgeasy.h>
#include <readline/readline.h>

typedef enum { FALSE, TRUE } bool;

typedef struct
{
  char  align;
  int   length;
} result_fld;

struct
{
  bool  show_metadata;
  int   text_len;
} options = { FALSE, 35 };

void dump_conn( PGconn * conn )
{
  const char * stats[] =
  {
    "CONNECTION_OK",
      "CONNECTION_BAD",
      "CONNECTION_STARTED",
      "CONNECTION_MADE",
      "CONNECTION_AWAITING_RESPONSE",
      "CONNECTION_AUTH_OK",
      "CONNECTION_SETENV"
  };
  
  if( PQstatus( conn ) == CONNECTION_OK )
  {
    printf( "\nConnected to database '%s' as user '%s'\n", 
            PQdb( conn ), 
            PQuser( conn ));
  }
  else
  {
    printf( "Connection status:\n" );
    printf( "dbname\t\t%s\n",  PQdb( conn ));
    printf( "user\t\t%s\n",    PQuser( conn ));
    printf( "pass\t\t%s\n",    PQpass( conn ));
    printf( "host\t\t%s\n",    PQhost( conn ));
    printf( "port\t\t%s\n",    PQport( conn ));
    printf( "tty\t\t%s\n",     PQtty( conn ));
    printf( "options\t%s\n",   PQoptions( conn ));
    printf( "error\t\t%s\n",   PQerrorMessage( conn ));
    printf( "backend\t\t%d\n", PQbackendPID( conn ));
    printf( "status\t\t%s\n",  stats[PQstatus( conn )] );
  } 
}

void dump_query( char * buf, PGresult * result )
{
  printf( "query\t\t%s\n",     buf );
  printf( "status\t\t%d\n",    PQresultStatus( result ));
  printf( "message\t\t%s\n",   PQresultErrorMessage( result ));
  printf( "tuples\t\t%d\n",    PQntuples( result ));
  printf( "cmdtuples\t\t%s\n", PQcmdTuples( result ));
}

static bool status_ok( PGresult * result )
{
  ExecStatusType  status = PQresultStatus( result );
  
  if( status == PGRES_TUPLES_OK || status == PGRES_COMMAND_OK )
    return( TRUE );
  else
    return( FALSE );
}

void process_results( PGresult * result )
{
  int          col;
  int          row;
  result_fld * fields;
  
  if( status_ok( result ) == FALSE )
  {
    fprintf( stderr, "%s\n", PQresultErrorMessage( result ));
  }
  else
  {
    fields = (result_fld *)calloc( PQnfields( result ), 
                                   sizeof( fields[0] ));
    
    if( options.show_metadata )
    {
      for( col = 0; col < PQnfields( result ); col++ )
      {
        printf( "%-3d: %-32s - type=%d, mod=%d, size=%d\n", 
          col, 
          PQfname( result, col ),
          PQftype( result, col ),
          PQfmod( result, col ),
          PQfsize( result, col ));
      }
    }
    
/*
**  Compute the field lengths
*/    
    for( col = 0; col < PQnfields( result ); col++ )
    {
      fields[col].align = '-';
      fields[col].length = strlen( PQfname( result, col ));
    }
    
    for( row = 0; row < PQntuples( result ); row++ )
    {
      for( col = 0; col < PQnfields( result ); col++ )
      {
        int field_len;
        
        if( PQgetisnull( result, row, col ))
          field_len = 1;
        else
          field_len = strlen( PQgetvalue( result, row, col ));
        
        if( field_len > fields[col].length )
          fields[col].length = field_len;
      }
    }
    
/*
**  Print the column headers
*/
    for( col = 0; col < PQnfields( result ); col++ )
    {
      printf( "%-*s ", fields[col].length, PQfname( result, col ));
    }
    
    printf( "\n" );
    
    for( col = 0; col < PQnfields( result ); col++ )
    { 
      int i;
      
      for( i = 0; i < fields[col].length; i++ )
        printf( "%c", '-' );
      
      printf( " " );
      
    }
    
    printf( "\n" );
    
    for( row = 0; row < PQntuples( result ); row++ )
    {
      for( col = 0; col < PQnfields( result ); col++ )
      {
        if( PQgetisnull( result, row, col ))
          printf( "%*s", fields[col].length, "" );
        else
        {
          if( fields[col].align == '-' )
          {
            printf( "%-*.*s", fields[col].length, 
                              fields[col].length, 
                              PQgetvalue( result, row, col ));
          }
          else
          {
            printf( "%*.*s", fields[col].length, 
                             fields[col].length, 
                             PQgetvalue( result, row, col ));
          }
        }
        printf(" ");
      }
      printf("\n");
    }
    
    free( fields );
  }
}

void process_query( char * buf )
{
  PGresult *  result;
  
  result = doquery( buf );
  
  if( options.show_metadata )
    dump_query( buf, result ); 
  
  process_results( result );
  
}

void process_internal_command( char * buf )
{
  if( strcmp( buf, "\\show metadata" ) == 0 )
  {
    options.show_metadata = TRUE;
    printf( "ok, showing metadata\n" );
  }
  else if( strcmp( buf, "\\hide metadata" ) == 0 )
  {
    options.show_metadata = FALSE;
    printf( "ok, hiding metadata\n" );
  }
  else if( strncmp( buf, "\\set text_length", 
          strlen( "\\set text_length" )) == 0 )
  {
    sscanf( buf, "%*s %*s %d", &options.text_len );
    printf( "ok, showing %d characters from text columns\n", 
            options.text_len );
  }
  else if( strcmp( buf, "\\on error stop" ) == 0 )
  {
    on_error_stop();
    printf( "ok, stopping on error\n" );
  }
  else if( strcmp( buf, "\\on error continue" ) == 0 )
  {
    on_error_continue();
    printf( "ok, continuing on error\n" );
  }
}

int main( int argc, char * argv[] )
{
  char    * buf;
  
  dump_conn( connectdb( argc > 1 ? argv[1] : "" ));
  
  on_error_continue();
  
  using_history();
  read_history( NULL );
  
  while(( buf = readline( "-->" )) != NULL )
  {
    if( strncmp( buf, "quit", sizeof( "quit" ) - 1  ) == 0 )
      break;
    else if( buf[0] == '\\' )
    {
      process_internal_command( buf );
    }
    else
    {
      add_history( buf );
      process_query( buf );
    }
  }
  
  write_history( NULL );
  
  disconnectdb();
  
}

