#!/usr/bin/perl -W
#
#  Filename: client4.pl
#

use DBI;
use Term::ReadLine;

my $dbh = DBI->connect("dbi:Pg:", undef, undef, {PrintError => 1})
  or die "Can't connect to PostgreSQL: $DBI::errstr ($DBI::err)\n";

my $term = new Term::ReadLine( 'client4' );

print( "\nEnter SQL commands or :help for assistance\n\n" );

while( my $command = $term->readline( "--> " )) {
    if( $command =~ /^:(\w+)\s*(.*)/i ) {
	eval {
	    my $code = "do_$1";
	    my @args = split '\s', $2||'';

	    &$code( $dbh, @args );
	}
    }
    else {
	do_sql_command( $dbh, $command );
    }
}

do_quit( $dbh );


#
# subroutine: do_sql_command( $dbh, $command )
#    Executes given SQL command and 
#    prints results
#
sub do_sql_command 
{
  my $dbh     = shift;
  my $command = shift;
  
  my $sth = $dbh->prepare( $command );

  if( defined( $sth )) {
      if( $sth->execute()) {
	  process_results( $dbh, $sth );
      }
  }
}

#
# subroutine: do_ping()
#    Tests connection to database
#
sub do_ping
{
  my( $dbh, @args ) = @_;

  print( $dbh->ping() ? "Ok\n" : "Not On" );
}

#
# subroutine: do_autocommit( [0|1] )
#    Enable or disable autocommit
#
sub do_autocommit
{
  my( $dbh, @args ) = @_;    

  $dbh->{AutoCommit} = $args[0];

}

#
# subroutine: do_commit()
#    Commits transaction block
#
sub do_commit
{
  my( $dbh, @args ) = @_;
  
  $dbh->commit();
}

#
# subroutine: do_rollback()
#    Aborts transaction block
#
sub do_rollback
{
    my( $dbh, @args ) = @_;

    $dbh->rollback();
}

#
# subroutine: do_quit()
#    Close connection and exit
#
sub do_quit
{
    my( $dbh, @args ) = @_;

    if( defined( $dbh )) {
	$dbh->disconnect();
    }

    exit( 0 );
}

#
# subroutine: do_trace( [0..4] )
#    Sets DBI trace level
#
sub do_trace
{
    my( $dbh, @args ) = @_;

    $dbh->trace( @args );

}

#
# subroutine: do_help()
#    Print help text
#
sub do_help
{
    print( "Commands\n" );
    print( "  :help\t\t\t\tShow help text\n" );
    print( "  :autocommit [0:1]\t\tSet AutoCommit\n" );
    print( "  :commit\t\t\tCOMMIT TRANSACTION\n" );
    print( "  :rollback\t\t\tROLLBACK TRANSACTION\n" );
    print( "  :trace [0|1|2|3|4]\t\tSet Trace level\n" );
    print( "  :show_tables\t\t\tShow all table names\n" );
    print( "  :show_table table_name\tDescribe table\n" );
    print( "  :show_types\t\t\tList Data Types\n" );
}

#
# subroutine: do_show_tables()
#    Display list of accessible tables
#
sub do_show_tables
{
  my( $dbh, @args ) = @_;

  process_results( $dbh, $dbh->table_info());

}

#
# subroutine: do_show_types()
#    Display list of data types
#
sub do_show_types
{
  my( $dbh, @args ) = @_;

  print("Type            Type         SQL  Col.  Prefix  \n");
  print("Name            Parameters   Type Size    Suffix\n");
  print("--------------- ------------ ---- ----- - ------\n" );

  foreach my $type ( $dbh->type_info( undef )) {
      printf( "%-15s %-12s %-3d  %-5d %s %s\n",
	      $type->{TYPE_NAME},
	      $type->{CREATE_PARAMS} || "",
	      $type->{DATA_TYPE},
	      $type->{COLUMN_SIZE},
	      $type->{LITERAL_PREFIX} || " ",
	      $type->{LITERAL_SUFFIX} || " " );
  }
}

#
# subroutine: do_show_table( $tablename )
#    Describe given table
#
sub do_show_table
{
  my( $dbh, @args ) = @_;

  my $sth = $dbh->prepare( "SELECT * FROM $args[0] WHERE 1 <> 1" );

  if( defined( $sth )) {
      if( $sth->execute()) {
	  print_meta_data( $dbh, $sth );
	  $sth->finish();
      }
  }
}

#
# subroutine: process_results( $dbh, $sth )
#    Print command results
#
sub process_results
{
  my $dbh = shift;
  my $sth = shift;

  if( defined( $sth )) {
      if( $sth->{NUM_OF_FIELDS} == 0 ) {
	print( $sth->{pg_cmd_status} . "\n" );
      }
      else {
        my($widths, $row_values) = compute_column_widths( $sth );
	print_column_headings( $sth, $widths );
	print_results( $sth, $row_values, $widths );
    }
  }
}

#
# subroutine: print_meta_data( $dbh, $sth )
#    Print command metadata
#
sub print_meta_data
{
  my $dbh = shift;
  my $sth = shift;

  my $field_count = $sth->{NUM_OF_FIELDS};
  my $names       = $sth->{NAME};
  my $pg_types    = $sth->{pg_type};

  print( "Name                          | Type   \n" );
  print( "------------------------------+--------\n" );

  for( my $col = 0; $col < $field_count; $col++ ) {
    printf( "%-30s| %-8s\n", $names->[$col], $pg_types->[$col] );
  }
}

#
# subroutine: compute_column_widths( $sth )
#    Compute width of each column in result set
sub compute_column_widths
{
  my $sth   = shift;
  my $names = $sth->{NAME};
  my @widths;

  for( my $col = 0; $col < $sth->{NUM_OF_FIELDS}; $col++ ) {
    push( @widths, length( $names->[$col] ));
  }

  my $row_values = $sth->fetchall_arrayref();

  for( my $col = 0; $col < $sth->{NUM_OF_FIELDS}; $col++ ) {
    for( my $row = 0; $row < $sth->rows(); $row++ ) {
      if( defined( $row_values->[$row][$col] )) {
        if( length( $row_values->[$row][$col] ) > $widths[$col] ) {
	  $widths[$col] = length( $row_values->[$row][$col] );
        }
      }
    }
  }

  return( \@widths, $row_values );
}

#
# subroutine: pad( $val, $col_width, $pad_char )
#    Pad given value to $col_width with (optional) $pad_char
#
sub pad
{
  my( $val, $col_width, $pad_char ) = @_;
  my $pad_len;

  $val      = "" if ( !defined( $val ));
  $pad_char = " " if( !defined( $pad_char ));
  $pad_len  = $col_width - length( $val );

  return( $val . $pad_char x $pad_len . " " );

}

#
# subroutine: print_column_headings( $sth, $widths )
#    Print column headings
#
sub print_column_headings
{
  my $sth    = shift;
  my $widths = shift;
  my $names  = $sth->{NAME};
    
  for( my $col = 0; $col < $sth->{NUM_OF_FIELDS}; $col++ ) {
    print( pad( $names->[$col], $widths->[$col] ));
  }

  print( "\n" );

  for( my $col = 0; $col < $sth->{NUM_OF_FIELDS}; $col++ ) {
    print( pad( "-", $widths->[$col], "-" ));
  }

  print( "\n" );
}

#
# subroutine: print_results( $sth, $row_values, $widths )
#    Print result set
#
sub print_results
{
  my( $sth, $rows, $widths ) = @_;

  for( my $row = 0; $row < $sth->rows(); $row++ ) {
    for( my $col = 0; $col < $sth->{NUM_OF_FIELDS}; $col++ ) {
      print( pad( $rows->[$row][$col], $widths->[$col] ));
    }
    print( "\n" );
  }
}
