/*
** File:  client4.c
*/

#include <stdlib.h>
#include <string.h>
#include <libpq-fe.h>
#include <readline/readline.h>
#include <readline/history.h>

#include <sys/time.h>

typedef enum { FALSE, TRUE } bool;

#define MAX_PRINT_LEN	40

static char separator[MAX_PRINT_LEN+1];

void print_result_set( PGresult * result )
{
  int          col;
  int          row;
  int	     * sizes;

/*
** Compute the size for each column
*/
  sizes = (int *)calloc( PQnfields( result ), sizeof( int ));
  
  for( col = 0; col < PQnfields( result ); col++ )
  {
    int	len = 0;

    for( row = 0; row < PQntuples( result ); row++ )
    {
      if( PQgetisnull( result, row, col ))
        len = 0;
      else
	len = PQgetlength( result, row, col );

      if( len > sizes[col] )
	sizes[col] = len;
    }

    if(( len = strlen( PQfname( result, col ))) > sizes[col] )
      sizes[col] = len;

    if( sizes[col] > MAX_PRINT_LEN )
      sizes[col] = MAX_PRINT_LEN;
  }

/*
** Print the field names.
*/
  for( col = 0; col < PQnfields( result ); col++ )
  {
    printf( "%-*s ", sizes[col], PQfname( result, col ));
  }

  printf( "\n" );

/*
** Print the separator line
*/
  memset( separator, '-', MAX_PRINT_LEN );

  for( col = 0; col < PQnfields( result ); col++ )
  {   
    printf( "%*.*s ", sizes[col], sizes[col], separator );
  }

  printf( "\n" );

/*
** Now loop through each of the tuples returned by
** our query and print the results.
*/
  for( row = 0; row < PQntuples( result ); row++ )
  {
    for( col = 0; col < PQnfields( result ); col++ )
    {
      if( PQgetisnull( result, row, col ))
	printf( "%*s", sizes[col], "" );
      else
	printf( "%*s ", sizes[col], PQgetvalue( result, row, col ));
    }

    printf( "\n" );

  }
  printf( "(%d rows)\n\n", PQntuples( result ));
  free( sizes );
}

bool is_result_ready( PGconn * connection )
{
    int			my_socket;
    struct timeval	timer;
    fd_set		read_mask;

    if( PQisBusy( connection ) == FALSE )
	return( TRUE );

    my_socket = PQsocket( connection );

    timer.tv_sec  = (time_t)1;
    timer.tv_usec = 0;
    
    FD_ZERO( &read_mask );
    FD_SET( my_socket, &read_mask );

    if( select( my_socket + 1, &read_mask, NULL, NULL, &timer ) == 0 )
    {
      return( FALSE );
    }
    else if( FD_ISSET( my_socket, &read_mask ))
    {
      PQconsumeInput( connection );

      if( PQisBusy( connection ))
	return( FALSE );
      else
	return( TRUE );
    }
    else
    {
      return( FALSE );
    }
}

void process_query( PGconn * connection, const char * query_text )
{
  PGresult	*	result;

  PQsetnonblocking( connection, 1 );

  if( PQsendQuery( connection, query_text ) == 0 )
  {
    printf( "%s\n", PQerrorMessage( connection ));
    return;
  }

  do
  {
    while( is_result_ready( connection ) == FALSE )
    {
      putchar( '.' );
      fflush( stdout );
    }
    printf( "\n" );

    if(( result = PQgetResult( connection )) != NULL )
    {
      if( PQresultStatus( result ) == PGRES_TUPLES_OK )
      {
	print_result_set( result );
      }
      else if( PQresultStatus( result ) == PGRES_COMMAND_OK )
      {    
	printf( "%s", PQcmdStatus( result ));

	if( strlen( PQcmdTuples( result )))
	  printf( " - %s rows\n", PQcmdTuples( result ));
	else
          printf( "\n" );
      }
      else
      {
	printf( "%s\n", PQresultErrorMessage( result ));
      }
      PQclear( result );
    }
  } while( result != NULL );
}

int main( int argc, char * argv[] )
{
  PGconn  * connection;
  char    * buf;
  int	    err;

  connection = PQconnectdb( argc > 1 ? argv[1] : "" );

  if( connection == NULL )
  {
    printf( "Fatal error - unable to allocate connection\n" );
    exit( EXIT_FAILURE );
  }

  if( PQstatus( connection ) == CONNECTION_OK )
  {
  
    using_history();
    err = read_history( ".pg_history" );
  
    while(( buf = readline( "-->" )) != NULL )
    {
      if( strncmp( buf, "quit", sizeof( "quit" ) - 1  ) == 0 )
      {
	break;
      }
      else
      {
	if(strlen( buf ) != 0 )
	{
	  add_history( buf );
	  process_query( connection, buf );
	}
	free( buf );
      }
    }

    err = write_history( ".pg_history" );

  }
  else
  {
    printf( "%s\n", PQerrorMessage( connection ));
  }
  
  PQfinish( connection );

  exit( EXIT_SUCCESS );
  
}

