/* qt-sql.cpp */

#include <qapplication.h>    // QT Basic application classes
#include <qwidget.h>         // QT Basic widget class
#include <qtable.h>          // QT Table widget
#include <qmessagebox.h>     // QT MessageBox widget
#include <qlayout.h>         // QT Layout manager
#include <qpushbutton.h>     // QT Pushbutton widget
#include <qmultilineedit.h>  // QT MultiLineEdit widget
#include <qstatusbar.h>      // QT Statusbar widget

#include <sql.h>             // ODBC Library
#include <sqlext.h>          // ODBC Extensions
#include <iostream.h>        // Standard C++ io library

#include <stdlib.h>
#include <malloc.h>


#include "qt-sql.moc"

typedef struct
{
  char          name[128+1];
  SQLSMALLINT   nameLength;
  SQLSMALLINT   dataType;
  SQLUINTEGER   fieldLength;
  SQLSMALLINT   scale;
  SQLSMALLINT   nullable;
  SQLINTEGER    displaySize;
  int           headerLength;
  SQLINTEGER    resultLength;
  char        * value;
} resultField;

void main( int argc, char * argv[] )
{

  QApplication a( argc, argv );
  MyMain       w( a.argv()[1] ? a.argv()[1] : "" );

  a.setMainWidget( &w );

  w.show();
  a.exec();
}

bool SQL_OK( SQLRETURN result )
{
  if( result == SQL_SUCCESS || result == SQL_SUCCESS_WITH_INFO )
    return( TRUE );
  else
    return( FALSE );
}

MyMain::MyMain( const char * connectStr )
{
  // Establish a reasonable size for our main window
  resize( 700, 550 );

  // Create two layout helpers -
  //   the vbox layout object will stack things vertically
  //   the buttons layout object will hold a row of buttons

  QVBoxLayout *  vbox    = new QVBoxLayout( this );
  QHBoxLayout *  buttons = new QHBoxLayout();

  //
  //  Create the user-interface components
  //
  edit   = new QMultiLineEdit( this );
  status = new QStatusBar( this );
  table  = new MyTable( this, connectStr );

  // remove the resize-grip from the statusbar,
  // it looks kinda strange in the middle of a 
  // window.

  status->setSizeGripEnabled( FALSE );
  status->message( "Connected" );

  //
  //  give the keyboard focus to the editor control
  //
  edit->setFocus();

  connect( edit, SIGNAL( returnPressed()), this, SLOT( execute()));

  vbox->addWidget( edit );
  vbox->addWidget( status );
  vbox->addWidget( table );

  // And finally create the row of buttons at
  // the bottom of the main window (quit, execute)
  //
  vbox->addLayout( buttons );
	
  QPushButton * quit = new QPushButton( "Quit", this );
  connect( quit, SIGNAL( clicked()), qApp, SLOT( quit()));

  QPushButton * exec = new QPushButton( "Execute", this );
  connect( exec, SIGNAL( clicked()), this, SLOT( execute()));

  buttons->addWidget( quit );
  buttons->addWidget( exec );
	
}

MyTable::MyTable( QWidget * parent, const char * connect ) 
  : QTable( parent )
{
  //
  //  Create a database connection...
  //
  SQLRETURN  res;

  res = SQLAllocHandle( SQL_HANDLE_ENV,
			 SQL_NULL_HANDLE,
			 &env );
  if( !SQL_OK( res ))
  {
    displayErrors( SQL_HANDLE_ENV, env );
    exit( -1 );
  }

  SQLSetEnvAttr( env, 
		 SQL_ATTR_ODBC_VERSION, 
		 (SQLPOINTER)SQL_OV_ODBC2, 
		 0 );

  res = SQLAllocHandle( SQL_HANDLE_DBC,
			env,
			&db );

  if( !SQL_OK( res ))
  {
    displayErrors( SQL_HANDLE_ENV, env );
    exit( -1 );
  }

  res = SQLConnect( db, 
		    (SQLCHAR *)qApp->argv()[1], SQL_NTS,
		    (SQLCHAR *)qApp->argv()[2], SQL_NTS,
		    (SQLCHAR *)qApp->argv()[3], SQL_NTS );
  
  if( !SQL_OK( res ))
  {
    displayErrors( SQL_HANDLE_DBC, db );
    exit( -1 );
  }


  //  We don't have any table-oriented results to 
  //  show yet, so hide the table.
  //
  setNumRows( 0 );
  setNumCols( 0 );
}

void MyTable::displayErrors( SQLSMALLINT type, SQLHANDLE handle )
{
  SQLHDBC   dbc  = SQL_NULL_HDBC;
  SQLHENV   env  = SQL_NULL_HENV;
  SQLHSTMT  stmt = SQL_NULL_HSTMT;

  switch( type )
  {
    case SQL_HANDLE_ENV:  env  = (SQLHENV)handle; break;
    case SQL_HANDLE_DBC:  dbc  = (SQLHENV)handle; break;
    case SQL_HANDLE_STMT: stmt = (SQLHSTMT)handle; break;
  }

  SQLRETURN   result;
  SQLCHAR     sqlState[6];
  SQLINTEGER  nativeError;
  SQLSMALLINT requiredLength;
  SQLCHAR     messageText[SQL_MAX_MESSAGE_LENGTH+1];

  QDialog     * dlg  = new QDialog( this, 0, TRUE );
  QVBoxLayout * vbox = new QVBoxLayout( dlg );
  QPushButton * ok   = new QPushButton( "Ok", dlg );

  setCaption( "Error" );
  QMultiLineEdit * edit = new QMultiLineEdit( dlg );

  vbox->addWidget( edit );
  vbox->addWidget( ok );

  connect( ok, SIGNAL( clicked()), dlg, SLOT( accept()));

  edit->setReadOnly( TRUE );

  do
  {
    result = SQLError( env,
                       dbc,
                       stmt,
                       sqlState,
                       &nativeError,
                       messageText,
                       sizeof( messageText ),
                       &requiredLength );

    if( SQL_OK( result ))
    {
      edit->append((char *)messageText );      
      edit->append( "\n" );
    }
  } while( SQL_OK( result ));

  dlg->adjustSize();
  dlg->exec();

}

void MyMain::execute( void )
{
  //  This function is called whenever the user
  //  presses the 'Execute' button (or whenever
  //  the user presses the Return key while the
  //  edit control has the keyboard focus)
  SQLHDBC   db = table->db;
  SQLHSTMT  stmt;
  SQLRETURN res;
  QString   qcmd = edit->text();
  SQLCHAR * cmd;

  // Convert the query command from Unicode
  // into an 8-bit, SQLCHAR format

  cmd = (SQLCHAR *)qcmd.latin1();

  SQLAllocHandle( SQL_HANDLE_STMT, db, &stmt );

  res = SQLPrepare( stmt, (SQLCHAR *)cmd, SQL_NTS );

  if( !SQL_OK( res ))
  {
    table->displayErrors( SQL_HANDLE_STMT, stmt );
  }
  else
  {

    if( SQL_OK( SQLExecute( stmt )))
    {
      SQLSMALLINT  columnCount;

      SQLNumResultCols( stmt, &columnCount );
    
      if( columnCount == 0 )
      {
	SQLINTEGER  rowCount;
	SQLRowCount( stmt, &rowCount );
	
	if( rowCount == -1 )
	  status->message( "Ok" );
	else
	{
	  QString m( "Ok, %1 rows affected" );

	  status->message( m.arg((int)rowCount ));
	}
      }
      else
      {
	status->message( "Ok..." );
	table->buildTable( stmt );
      }
    }
    else
      table->displayErrors( SQL_HANDLE_STMT, stmt );

  }

  SQLFreeHandle( SQL_HANDLE_STMT, stmt );
}

void MyTable::buildTable( SQLHSTMT stmt )
{
  //  This function is called to fill in 
  //  the table control.  We want to fill
  //  the table with the result set.
  SQLSMALLINT   i;
  SQLSMALLINT   columnCount;
  resultField * fields;

  setNumRows( 0 );
  setNumCols( 0 );

  //  First, examine the metadata for the 
  //  result set so that we know how much
  //  room we need for each column.

  SQLNumResultCols( stmt, &columnCount );

  fields = new resultField[ columnCount+1 ];

  setNumCols( columnCount );

  for( i = 1; i <= columnCount; i++ )
  {
    SQLDescribeCol( stmt,
		    i,
		    (SQLCHAR *)fields[i].name,
		    sizeof( fields[i].name ),
		    &fields[i].nameLength,
		    &fields[i].dataType,
		    &fields[i].fieldLength,
		    &fields[i].scale,
		    &fields[i].nullable );

    SQLColAttribute( stmt, 
		     i,
		     SQL_DESC_DISPLAY_SIZE,
		     NULL,
		     0,
		     NULL,
		     &fields[i].displaySize );

    fields[i].value = (char *)malloc( fields[i].displaySize+1 );

    // Build the column headers as we go
    horizontalHeader()->setLabel( i-1, fields[i].name );

  }

  //  Bind the fields to our buffers
  for( i = 1; i <= columnCount; i++ )
  {
    SQLRETURN res;

    res = SQLBindCol( stmt,
		      i,
		      SQL_C_CHAR,
		      fields[i].value,
		      fields[i].displaySize+1,
		      &fields[i].resultLength );

    if( !SQL_OK( res ))
      displayErrors( SQL_HANDLE_STMT, stmt );
  }

  //
  //  Now, put the data into the table...
  //
  int       row = 0;
  SQLRETURN res;

  while( SQL_OK(( res =  SQLFetch( stmt ))))
  {
    if( res == SQL_SUCCESS_WITH_INFO )
      displayErrors( SQL_HANDLE_STMT, stmt );

    setNumRows( row+1 );

    for( int col = 1; col <= columnCount; col++ )
    {
      setText( row, col-1, fields[col].value );
    }

    row++;

  }
}




