/* qt-sql.cpp */

#include <qapplication.h>    // QT Basic application classes
#include <qwidget.h>         // QT Basic widget class
#include <qtable.h>          // QT Table widget
#include <qmessagebox.h>     // QT MessageBox widget
#include <qlayout.h>         // QT Layout manager
#include <qpushbutton.h>     // QT Pushbutton widget
#include <qmultilineedit.h>  // QT MultiLineEdit widget
#include <qstatusbar.h>      // QT Statusbar widget

#include <libpq++.h>         // PostgreSQL libpq++ API 
#include <iostream.h>        // Standard C++ io library

#include "qt-sql.moc"

void main( int argc, char * argv[] )
{

  QApplication a( argc, argv );
  MyMain       w( a.argv()[1] ? a.argv()[1] : "" );

  a.setMainWidget( &w );

  w.show();
  a.exec();
}

MyMain::MyMain( const char * connectStr )
{
  // Establish a reasonable size for our main window
  resize( 640, 450 );

  // Create two layout helpers -
  //   the vbox layout object will stack things vertically
  //   the buttons layout object will hold a row of buttons

  QVBoxLayout *  vbox    = new QVBoxLayout( this );
  QHBoxLayout *  buttons = new QHBoxLayout();

  //
  //  Create the user-interface components
  //
  edit   = new QMultiLineEdit( this );
  status = new QStatusBar( this );
  table  = new MyTable( this, connectStr );

  // remove the resize-grip from the statusbar,
  // it looks kinda strange in the middle of a 
  // window.

  status->setSizeGripEnabled( FALSE );

  //
  //  give the keyboard focus to the editor control
  //
  edit->setFocus();

  connect( edit, SIGNAL( returnPressed()), this, SLOT( execute()));

  vbox->addWidget( edit );
  vbox->addWidget( status );
  vbox->addWidget( table );

  // And finally create the row of buttons at
  // the bottom of the main window (quit, execute)
  //
  vbox->addLayout( buttons );
	
  QPushButton * quit = new QPushButton( "Quit", this );
  connect( quit, SIGNAL( clicked()), this, SLOT( quit()));

  QPushButton * exec = new QPushButton( "Execute", this );
  connect( exec, SIGNAL( clicked()), this, SLOT( execute()));

  buttons->addWidget( quit );
  buttons->addWidget( exec );
	
}

MyTable::MyTable( QWidget * parent, const char * connect ) 
  : QTable( parent )
{
  //
  //  Create a database connection...
  //
  db = new PgDatabase( connect );

  if( db->ConnectionBad())
  {
    QMessageBox::critical( 0, "Connection failed", 
			   db->ErrorMessage());
    exit( -1 );
  }

  //  We don't have any table-oriented results to 
  //  show yet, so hide the table.
  //
  setNumRows( 0 );
  setNumCols( 0 );
}

void MyMain::execute( void )
{
  //  This function is called whenever the user
  //  presses the 'Execute' button (or whenever
  //  the user presses the Return key while the
  //  edit control has the keyboard focus)

  PgDatabase    * db = table->db;
  ExecStatusType  result;

  //  Execute whatever the user has entered into
  //  the edit control
  //
  result = db->Exec(( const char *)edit->text());

  //
  //  Now process the results...
  //
  switch( result )
  {
    case PGRES_EMPTY_QUERY:
      status->message( "That was fun..." );
      break;

    case PGRES_COMMAND_OK:
      status->message( "Ok" );
      break;

    case PGRES_TUPLES_OK:
      status->message( "Ok..." );
      table->buildTable();
      break;

    default:
      status->message( db->ErrorMessage());
      break;
  }
}

void MyTable::buildTable( void )
{
  //  This function is called to fill in 
  //  the table control.  We want to fill
  //  the table with the result set.

  setNumRows( db->Tuples());
  setNumCols( db->Fields());

  //
  //  First, populate the column headers...
  //
  for( int col = 0; col < db->Fields(); col++ )
  {
      horizontalHeader()->setLabel( col, db->FieldName( col ));
  }

  //
  //  Now, put the data into the table...
  //
  for( int row = 0; row < db->Tuples(); row++ )
  {
    for( int col = 0; col < db->Fields(); col++ )
    {
      setText( row, col, db->GetValue( row, col ));
    }
  }
}

void MyMain::quit( void )
{
  PgDatabase * db = table->db;

  if( db != NULL )
  {
    delete db;
    db = NULL;
  }

  qApp->quit();

}




